//
//								XYZPLOT
// 2-D visualization program for ASCII data stored in three text columns: X, Y, and Z
// (plots colored 2D maps using the Z-values as input for color scheme)  
//
// Copyright 2011-2015 by Yuri Ganshin, CMI, University of Wyoming, Laramie.
// http://www.uwyo.edu/cmi/
//
// This program is free sofware released using the Fast Light Tool Kit (FLTK).
// For conditions of distribution and use, see the license agreement in 
//     http://www.fltk.org/COPYING.php
// As a general rule, "free software" means that the user have the freedom
// to run, copy, distribute, study, modify and improve the software.
//
// Please report all bugs and problems on the following page:
// http://www.uwyo.edu/cmi/
//     
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
//
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Scroll.H>
#include <FL/fl_draw.H>	
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Native_File_Chooser.H>
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Spinner.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_Float_Input.H>
#include <FL/Fl_Input_Choice.H>
#include <FL/Fl_Check_Button.H>
#include <FL/fl_show_colormap.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Box.H>
#include <FL/fl_message.H>
#include <FL/fl_ask.H> // for fl_alert() and fl_message()
#include <math.h>

using namespace std;

#define MARGIN		80
#define FLT_MAX	1E+37
#define FLT_MIN	-1E+37
#define PALSZ		11
#define PALINC		0.1	// Increment of values within a palette
//
////////////////////////////////////////////////////////////////
/****               List of GLOBAL variables               ****/
FILE *in_fp = NULL;
char inpname[200],c_word[4],cgbuf[90];
int chosen=-1, intercol=1;	
int Ncur = 1;				// Number of curves in file (should be 3)
int Ndat = 1;				// Number of data samples
int nx,ny,nempty, pt_size=1;
int win_w, win_h, canvas_w = 450, canvas_h = 900;	// pixels
//
float xticks = 1000.0, yticks = 1000.0;
float *zdat;	
float x_min,x_max,x_inc;
float y_min,y_max,y_inc;
float z_min,z_max,Zlow,Zhigh;
//
const char *Xtext_hdr = "  Cross-line numbers ";
const char *Ytext_hdr = "  In-line numbers    ";
//
///////////////////  Color Palettes  ////////////////////
// array size = PALSZ*4,  palette size = PALSZ == 11
//  All these palettes have range from 0.0 to 1.0
//
float *pal;				// Palette container
//
float RGB_pal[] =		// Red-Green-Blue 
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	125, 255, 255, 255, 125,   0,   0,   0,   0,   0,   0,
	  0,   0, 125, 255, 255, 175, 255, 255, 125,   0,   0,
	  0,   0,   0,   0,   0,   0, 125, 255, 255, 255, 125};
float BGR_pal[] =		// Blue-Green-Red
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	  0,   0,   0,   0,   0,   0, 125, 255, 255, 255, 125,
	  0,   0, 125, 255, 255, 175, 255, 255, 125,   0,   0,
	125, 255, 255, 255, 125,   0,   0,   0,   0,   0,   0};
//
float RGM_pal[] =		// Red-Green-Magenta
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	142, 255, 255, 255,   0,   0,   0,   0,  85, 139,  75,
	 35,   0, 127, 255, 255, 139, 245,   0,  26,   0,   0,
	 35,   0,   0,   0,   0,  69, 255, 205, 139, 139,  75};
float MGR_pal[] =		// Magenta-Green-Red
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	 75, 139,  85,   0,   0,   0,   0, 255, 255, 255, 142,
	  0,   0,  26,   0, 245, 139, 255, 255, 127,   0,  35,
	 75, 139, 139, 205, 255,  69,   0,   0,   0,   0,  35};
//
float BWR_pal[] =		// Blue-White-Red
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	  0,   0,   0,   0, 125, 255, 255, 255, 255, 255, 125,
	  0,   0, 125, 255, 255, 255, 255, 255, 125,   0,   0,
	125, 255, 255, 255, 255, 255, 125,   0,   0,   0,   0};
float RWB_pal[] =		// Red-White-Blue
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	125, 255, 255, 255, 255, 255, 125,   0,   0,   0,   0,
	  0,   0, 125, 255, 255, 255, 255, 255, 125,   0,   0,
	  0,   0,   0,   0, 125, 255, 255, 255, 255, 255, 125};
//
float WRB_pal[] =		// White-Red-Black
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	255, 255, 255, 255, 255, 255, 200, 150, 100,  50,   0,
	255, 255, 255, 255, 125,   0,   0,   0,   0,   0,   0,
	255, 170,  85,   0,   0,   0,   0,   0,   0,   0,   0};
float BRW_pal[] =		// Black-Red-White
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	  0,  50, 100, 150, 200, 255, 255, 255, 255, 255, 255,
	  0,   0,   0,   0,   0,   0, 125, 255, 255, 255, 255,
	  0,   0,   0,   0,   0,   0,   0,   0,  85, 170, 255};
//
float WGB_pal[] =		// White-Gray-Black
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	255, 220, 190, 175, 150, 125, 100,  75,  50,  25,   0,
	255, 220, 190, 175, 150, 125, 100,  75,  50,  25,   0,
    255, 220, 190, 175, 150, 125, 100,  75,  50,  25,   0};
float BGW_pal[] =		// Black-Gray-White
{	0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0,
	  0,  25,  50,  75, 100, 125, 150, 175, 190, 220, 255,
	  0,  25,  50,  75, 100, 125, 150, 175, 190, 220, 255,
      0,  25,  50,  75, 100, 125, 150, 175, 190, 220, 255};
//
////////////////////////////////////////////////////////////////
/****            List of Function Prototypes               ****/

void read_file_info(const char *name);
void pal_interp(float [], float, short []);
void pal_discr(float [], float, short []);

////////////////////////////////////////////////////////////////
/****         Global pointers to the GUI objects           ****/
////////////////////////////////////////////////////////////////

Fl_Double_Window *draw_window;
Fl_Window *opts_window;
Fl_Scroll *scroll;
Fl_Float_Input *min_x_inp, *max_x_inp, *min_y_inp, *max_y_inp;
Fl_Float_Input *xtick_inp, *ytick_inp, *lowZ_inp, *highZ_inp;
Fl_Button *apply, *reapply, *end_this;
Fl_Input *Xaxis_hdr, *Yaxis_hdr;
Fl_Check_Button *interpol;
//
/////////////////////////////////////////////////////////////////////
/****          List of Widgets and Drawing functions            ****/
/////////////////////////////////////////////////////////////////////

/**********     File-choosing Menu    ************/
class File_menu : public Fl_Window {
	Fl_Native_File_Chooser *fc;
	
	void read_data(const char *filename)	// Process the chosen file
	{
		strcpy(inpname,filename);
		chosen=1;			// will wait while(chosen<0)
	}
	
	// Handle an 'Open' request from the menu
	static void open_cb(Fl_Widget *w, void *v) {
		File_menu *choice = (File_menu*)v;
		choice->fc->title("Open");
		choice->fc->type(Fl_Native_File_Chooser::BROWSE_FILE);
		switch ( choice->fc->show() ) {
			case -1: break;	// Error
			case  1: break; 	// Cancel
			default:		// choice->fc->show() returns 0
				choice->read_data(choice->fc->filename());
				// choice->deactivate();				
				break;
		}
	}
	
	static void quit_cb(Fl_Widget *w, void *v) {
		exit(0);
	}
	
public:
	File_menu() : Fl_Window(500,300,350,150,"Choose the file") {
		Fl_Menu_Bar *menu = new Fl_Menu_Bar(0,0,350,25);
		menu->add("&File/&Open",  FL_COMMAND+'o', open_cb, (void*)this);
		menu->add("&File/&Quit",  FL_COMMAND+'q', quit_cb);
		
		// Describe the chooser
		Fl_Box *box = new Fl_Box(20,45,w()-40,h()-65);
		box->color(FL_GRAY-1); 
		box->box(FL_FRAME_BOX);
		box->align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE|FL_ALIGN_WRAP);
		box->label("Choose ASCII-format file with\n\
numbers\n\
arranged in columns and rows");
		box->labelsize(18);
		// Initialize the file chooser
		fc = new Fl_Native_File_Chooser();
		fc->filter("ASCII text files \t*.txt, *.TXT\n");			 
		end();
	}
};
/////////////////////////////////////////////////////////////////////
/////////		DATA  DRAWING in a scrollable window		/////////
/////////		Consider x() and y() values!!!				/////////
/////////////////////////////////////////////////////////////////////
// The Fl_Group class is the FLTK container widget, it maintains
// an array of child widgets. Fl_Window is a subclass of Fl_Group.
//
class Draw_data : public Fl_Group // Fl_Widget
{
	
public:
  	Draw_data(int X,int Y,int W,int H) : Fl_Group(X,Y,W,H) {}
	
	int handle(int event)
   {
		int ret = Fl_Group::handle(event);
		switch (event)
      {
			case FL_ENTER: // the mouse pointer entered a widget
				ret = 1;    // FL_ENTER: must return(1) to receive FL_PUSH
				break;
         case FL_PUSH:  // will cause 'user damage' with a specific code 
            if(Fl::event_button()==FL_LEFT_MOUSE) break;
            damage(FL_DAMAGE_USER1);
				ret = 1;    // FL_ENTER: must return(1) to receive FL_PUSH
				break;
		}
		return(ret);
	}
   //  damage(mask) indicates that a partial update of the object is needed
   
	void draw_rect() // Draw a circle at mouse pointer location
   {      
      int xx = (int)Fl::event_x();
      int yy = (int)Fl::event_y();
		int x_dim = w() - 2*MARGIN;  
		int y_dim = h() - 2*MARGIN;
		float iln,xln;
		
		fl_rectf(xx-1,yy-1,3,3,0,0,0); // fill a rectangle
		
		xln = x_min + (x_max-x_min)*(xx-x()-MARGIN)/x_dim;
		iln = y_min + (y_max-y_min)*(yy-y()-MARGIN)/y_dim;
		cout << "XLN = " << xln << ", ILN = " << iln << endl;		
	}

	void draw(void)  // Function draw() within class Draw_data
	{
		// User damage ONLY? - just draw a circle and done
		if ( damage() == FL_DAMAGE_USER1 ) { draw_rect(); return; }
	
		//      don't declare reserved 'x' and 'y' variables !!!
		int i,j;
		int xx, yy, x_sta, x_fin, x_dim, y_dim;
		short rgb[3];
		float val,sta;

		x_dim = w() - 2*MARGIN;  y_dim = h() - 2*MARGIN;

		fl_rectf(x(), y(), w(), h(), 235,235,235);
		fl_color(FL_BLACK);	fl_line_style(0,1); 
		fl_rect(x(), y(), w(), h());
		
		xx = x()+MARGIN;  yy = y()+MARGIN;
		fl_rect(xx-1, yy-1, x_dim+2, y_dim+2);				// Black outline
		fl_color(FL_WHITE);	fl_rectf(xx,yy,x_dim,y_dim);	// Work area

		///////////////////////////////////////////////////////////////////
		////  Fill the work area with color-coded normalized Z-values  ////
		///////////////////////////////////////////////////////////////////		
		for(j=0;j<ny;j++)	
		{
			yy = y()+MARGIN+j*pt_size;
			for(i=0;i<nx;i++)
			{
				xx = x()+MARGIN+i*pt_size;
				val = (zdat[nx*j+i]-Zlow)/(Zhigh-Zlow);
				if(val>1.0) val = 1.0;
				if(val<0.0) val = 0.0;
			
				if(intercol==1) pal_interp(pal,val,rgb);
				else pal_discr(pal,val+0.05,rgb);
				fl_rectf(xx,yy,pt_size,pt_size,rgb[0],rgb[1],rgb[2]);
				
			}
		}
		fl_color(FL_BLACK);	fl_font(FL_HELVETICA,12);	// (font,pixel_size)
		// Choices: FL_HELVETICA,  FL_COURIER,  FL_TIMES,  FL_SCREEN
		
		///////////////////////////////////////////////////////////////////
		//  Draw tick marks and labels on the X-axis
		
		i = (int)(x_min/xticks);  sta = xticks*(i+1); 
		for(val=sta; val<x_max; val+=xticks)
		{
			xx = x()+ MARGIN + (int)((val-x_min)*x_dim/(x_max-x_min));	
			yy = y() + MARGIN;
			// fl_color(23); fl_yxline(xx, yy, yy+y_dim);	// gray vertical line
			fl_color(0); fl_yxline(xx, yy-6, yy-1);			// vertical line
			sprintf(cgbuf,"%g",val);	fl_draw(cgbuf,xx-10,y()+MARGIN-12);
		}
		
		//  Draw tick marks and labels on the Y-axis
		i = (int)(y_min/yticks);  sta = yticks*(i+1);	
		for(val=sta; val<y_max; val+=yticks)
		{
			yy = y() + MARGIN + (int)((val-y_min)*y_dim/(y_max-y_min));
			//fl_color(23); fl_xyline(MARGIN, yy, w()); // gray horiz line
			fl_color(0); fl_xyline(MARGIN-6, yy, MARGIN-1);	// horiz line		
			sprintf(cgbuf,"%g",val);	fl_draw(cgbuf,40,yy+4);
		}
		
		///////////////////////////////////////////////////////////////////
		////  Drawing Color Palette

		x_sta = x()+ MARGIN + x_dim/5;
		yy = y()+ MARGIN + y_dim + 10;
		for(i=0,xx=x_sta; i<160; i++,xx++)
		{
			val=pal[0]+i*(pal[PALSZ-1]-pal[0])/160;
			if(intercol==1) pal_interp(pal, val ,rgb);
			else pal_discr(pal, val+0.05 ,rgb);
			fl_rectf(xx,yy,1,18,rgb[0],rgb[1],rgb[2]);
		}
		
		fl_color(FL_BLACK); fl_font(FL_HELVETICA,12); fl_line_style(0,1);
		yy = y()+ MARGIN + y_dim + 28;		//  Label Color Map
		for(i=0;i<5;i++)
		{	
			xx = x_sta + i*40;
			val = Zlow + (Zhigh-Zlow)*i/4.0;
			fl_yxline(xx, yy, yy+5);			// vertical line	
			sprintf(cgbuf,"%g",val);
			if(i%2==0) fl_draw(cgbuf,xx-12,yy+20);
			//else fl_draw(cgbuf,xx-8,yy+30);
		}
		fl_rect(x_sta, y()+MARGIN+y_dim+10, 161, 18); // Palette outline
		
		///////////////////////////////////////////////////////////////////
		//  Draw text headers along the axes
		
		xx = x_sta;
		yy = y()+ MARGIN + y_dim/2;		 
		fl_font(1,13);  fl_draw(Xtext_hdr,xx,40);
		//  Rotates the text string 90˚ counter-clockwise
		fl_draw(90,Ytext_hdr,25,yy);
		fl_draw("Color-coded Z-values",xx+20,y()+h()-16);

	} 	// void draw(void)	
	
} *data_plot;
//
//////////////////////////////////////////////////////////////
//
class Draw_pal : public Fl_Widget  ////  Drawing Color Palette
{
	void draw() 
	{
		int i,j,xx,yy,x_sta,wid,hei;
		short rgb[3];
		float yi;
		
		x_sta = x();  yy = y();  wid = w();  hei = h();
		fl_rectf(x_sta-2,yy-2,wid+4,hei+4,0,0,0); // Black frame
						
		for(i=0,xx=x_sta;i<wid;i++,xx++)
		{	
			yi=pal[0]+i*(pal[PALSZ-1]-pal[0])/wid;	// actually, yi=i/wid
			if(intercol==1) pal_interp(pal, yi ,rgb);
			else pal_discr(pal, yi+0.05 ,rgb);
			fl_rectf(xx,yy,1,hei,rgb[0], rgb[1], rgb[2]);
		}
	} 
	
public:
	Draw_pal(int X,int Y,int W,int H) : Fl_Widget(X,Y,W,H) {}
	
} *palette;
//
////////////////////////////////////////////////////////////////
/************    List of Call Back functions    ***************/
////////////////////////////////////////////////////////////////
//
void input_xtick(Fl_Widget *, void *) {
	xticks = atof(xtick_inp->value());
}
void input_ytick(Fl_Widget *, void *) {
	yticks = atof(ytick_inp->value());
}
//
void input_Zlow(Fl_Widget *, void *) {
	Zlow = atof(lowZ_inp->value());
}
void input_Zhigh(Fl_Widget *, void *) {
	Zhigh = atof(highZ_inp->value());
}
//
void spinner_cb(Fl_Widget *w, void *) {
   Fl_Spinner *s = (Fl_Spinner *)w;	
   pt_size = (int)s->value();
}
//
void Xhdr_cb(Fl_Widget *w, void *) {
   Fl_Input *s = (Fl_Input *)w;	
   Xtext_hdr = s->value();
}
void Yhdr_cb(Fl_Widget *w, void *) {
   Fl_Input *s = (Fl_Input *)w;	
   Ytext_hdr = s->value();
}
//
void choice_pal(Fl_Widget*,void*data) 
{
	Fl_Input_Choice *in_pal=(Fl_Input_Choice *)data;
	
	int i,num;
	
	strncpy(c_word,in_pal->value(),3);
	sscanf(c_word,"%d", &num);
	switch(num)
	{  
		case 0:	pal = RGB_pal; break;		
		case 1:	pal = BGR_pal; break;
		case 2:	pal = RGM_pal; break;			
		case 3:	pal = MGR_pal; break;			
		case 4:	pal = BWR_pal; break;			
		case 5:	pal = RWB_pal; break;
		case 6:	pal = WRB_pal; break;			
		case 7:	pal = BRW_pal; break;		
		case 8:	pal = WGB_pal; break;			
		case 9:	pal = BGW_pal; break;					
		default: break;
	}	
	palette->redraw();	
}
//
void inter_col(Fl_Widget *w, void *) {
   Fl_Check_Button *check = (Fl_Check_Button *)w;
   intercol = check->value();
	palette->redraw();
}
/////////////////////////////////////////////////////////////////////
///////////////           DRAWING Window          ///////////////////
/////////////////////////////////////////////////////////////////////

void doPLOT(Fl_Widget *, void*)
{  
	win_w = nx*pt_size + 2*MARGIN;	// Make left and right margins 
	win_h = ny*pt_size + 2*MARGIN;	// Make top and bottom margins
	
	draw_window = new Fl_Double_Window(400,400,"Data Plot Window");
	
	scroll = new Fl_Scroll(0,0,400,400);
	data_plot = new Draw_data(0,0,win_w,win_h);
	scroll->end();
	
	draw_window->end();
	draw_window->resizable(draw_window);	
	draw_window->show();

}
// Redraw data_plot with updated data
void doRedraw(Fl_Widget *, void*)
{	
	data_plot->redraw();
}
///////////    EXIT    ////////////
void exit_b(Fl_Widget *, void *) {
	exit(0);
}

/////////////////////////////////////////////////////////////////////
/****   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   ****/
/////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
	int i,cur;
	double t,val;
	char cbuf1[99],cbuf2[99],cbuf3[99],cbuf4[99],cbuf5[99],cbuf6[99];
	char cbuf[90];	
	////////////////////////////////////////////////////////////////
	
	Fl::scheme(NULL); //  "plastic" for Aqua and "gtk+"
	// NULL use the scheme defined in the FLTK_SCHEME env.var. 
	
	if(argc == 2) strcpy(inpname,argv[1]);		// Do not use terminal
	else		// if you want to create an Apple  *.app bundle application
	{
		File_menu *choice = new File_menu();
		choice->show();
		choice->end(); // no more widgets will be added to the window
		while(chosen<0) Fl::wait();
		choice->hide();
	}

	read_file_info(inpname);
	/////////////////////////////////////////////////////////////////
	
	opts_window = new Fl_Window(1050,50,420,800,"Display Manager");
	// The follwing widgets belong to this window	

	Fl_Box* label1 = new Fl_Box(FL_FRAME_BOX,20,10,380,180,""); label1->color(6);
	
	Fl_Box* label2 = new Fl_Box(80,20,250,20,"  File Information  ");
	label2->labelcolor(FL_BLACK);
	label2->labelfont(FL_HELVETICA_BOLD);	label2->labelsize(16);
		
	sprintf(cbuf1,"Input file contains %d triplets of XYZ data",Ndat);
	Fl_Box* label3 = new Fl_Box(80,45,250,20,cbuf1);
	label3->labelfont(FL_HELVETICA);	label3->labelsize(14);

	sprintf(cbuf2,"Number of data points along the X-axis = %d\n\
Number of data points along the Y-axis = %d",nx,ny);
	Fl_Box* label4 = new Fl_Box(80,70,250,20,cbuf2);
	
	sprintf(cbuf6,"%d empty (NULL) Z-values added to fill a rectangle.",nempty);
	Fl_Box* label8 = new Fl_Box(80,98,250,20,cbuf6);
	label8->labelfont(FL_HELVETICA);	label8->labelsize(14);	
	
	sprintf(cbuf3,"Range of X values = %g - %g  (inc = %g)",x_min,x_max,x_inc);
	Fl_Box* label5 = new Fl_Box(80,125,250,20,cbuf3);		

	sprintf(cbuf4,"Range of Y values = %g - %g  (inc = %g)",y_min,y_max,y_inc);
	Fl_Box* label6 = new Fl_Box(80,142,250,20,cbuf4);		
	
	sprintf(cbuf5,"Range of Z values = %g - %g",z_min,z_max);
	Fl_Box* label7 = new Fl_Box(80,159,250,20,cbuf5);

	/////////////////////////////////////////////////////////////////
		
	Fl_Box* label9 = new Fl_Box(70,200,240,30,"Select plot parameters:");
	label9->labelfont(FL_HELVETICA_BOLD); label9->labelsize(15);
	
	Fl_Spinner* spin = new Fl_Spinner(125,240,50,25,
												 "Point Size ");
	spin->labelsize(15);	spin->labelfont(FL_HELVETICA_BOLD);
	spin->minimum(1);		spin->maximum(5);		spin->step(1);
	spin->value(pt_size);	spin->align(Fl_Align(FL_ALIGN_LEFT));
	spin->callback(spinner_cb);
	Fl_Box* label14 = new Fl_Box(170,243,60,20,"pixel");
	label14->labelfont(FL_HELVETICA); label14->labelsize(15);	
	

	xtick_inp = new Fl_Float_Input(335, 275, 60, 25,
							"Interval between tick marks on the X-axis :");
	xtick_inp->labelfont(FL_HELVETICA_BOLD);	xtick_inp->labelsize(15);
	sprintf(cbuf," %g",xticks);					xtick_inp->value(cbuf);
	xtick_inp->callback(input_xtick);
	
	ytick_inp = new Fl_Float_Input(335, 315, 60, 25,
							"Interval between tick marks on the Y-axis :");
	ytick_inp->labelfont(FL_HELVETICA_BOLD);	ytick_inp->labelsize(15);
	sprintf(cbuf," %g",yticks);					ytick_inp->value(cbuf);
	ytick_inp->callback(input_ytick);
	//////////////////////////////////////////////////////////////////
	
	Fl_Input* Xaxis_hdr = new Fl_Input(170,355,200,25,
							"Label the X-axis: ");
	Xaxis_hdr->value(Xtext_hdr);				Xaxis_hdr->callback(Xhdr_cb);
	Xaxis_hdr->labelfont(FL_HELVETICA_BOLD);	Xaxis_hdr->labelsize(15);
	
	Fl_Input* Yaxis_hdr = new Fl_Input(170,390,200,25,
												"Label the Y-axis: ");
	Yaxis_hdr->value(Ytext_hdr);				Yaxis_hdr->callback(Yhdr_cb);
	Yaxis_hdr->labelfont(FL_HELVETICA_BOLD);	Yaxis_hdr->labelsize(15);
	
	//////////////////////////////////////////////////////////////////
	/////////////////          COLOR MAP            //////////////////
	//////////////////////////////////////////////////////////////////
	Fl_Input_Choice in_pal(185,440,200,25,"Choose colormap : ");
	in_pal.callback(choice_pal, (void*)&in_pal);
	
	in_pal.add("  0 - Red-Green-Blue   ");
	in_pal.add("  1 - Blue-Green-Red   ");
	in_pal.add("  2 - Red-Green-Magenta");
	in_pal.add("  3 - Magenta-Green-Red");
	in_pal.add("  4 - Blue-White-Red   ");
	in_pal.add("  5 - Red-White-Blue   ");
	in_pal.add("  6 - White-Red-Black ");
	in_pal.add("  7 - Black-Red-White ");
	in_pal.add("  8 - White-Gray-Black ");
	in_pal.add("  9 - Black-Gray-White ");	

	in_pal.value(0);  // default palette number
	in_pal.labelsize(15);
	in_pal.labelfont(FL_HELVETICA_BOLD);
	in_pal.align(Fl_Align(FL_ALIGN_LEFT));
	
	palette = new Draw_pal(90,495,232,16);	// 10*23+2 pixels  815
	
	Fl_Box* pal1 = new Fl_Box(80,473,250,20,
						"@<-  LOW          Z-values         HIGH  @->");
	pal1->labelfont(FL_HELVETICA_BOLD);	pal1->labelsize(12);
	
	
	interpol = new Fl_Check_Button(90,520,150,25,
									" Interpolate colormap colors?");
	interpol->labelfont(FL_HELVETICA_BOLD);	interpol->labelsize(15);
	interpol->value(intercol);			interpol->callback(inter_col);	
	
	
	//////////////////////////////////////////////////////////////////
	
	Fl_Box* pal2 = new Fl_Box(80,565,250,20,"Map Z-values into the colormap colors :");
	pal2->labelfont(FL_HELVETICA_BOLD);	pal2->labelsize(15);				

	lowZ_inp = new Fl_Float_Input(300, 590, 70, 25,
								"Z-value matching the leftmost color :  ");
	lowZ_inp->labelfont(FL_HELVETICA_BOLD);	lowZ_inp->labelsize(13);
	sprintf(cbuf," %g",Zlow);					lowZ_inp->value(cbuf);
	lowZ_inp->callback(input_Zlow);
	
	highZ_inp = new Fl_Float_Input(300, 625, 70, 25,
								"Z-value matching the rightmost color : ");
	highZ_inp->labelfont(FL_HELVETICA_BOLD);	highZ_inp->labelsize(13);
	sprintf(cbuf," %g",Zhigh);					highZ_inp->value(cbuf);
	highZ_inp->callback(input_Zhigh);							
													
	//////////////////////////////////////////////////////////////////
						
	apply = new Fl_Button(100,685,140,30,"Plot the data");
	apply->color(62);				apply->labelcolor(FL_BLACK);
	apply->labelfont(3);			apply->labelsize(16);
	apply->callback(doPLOT);
	
	reapply = new Fl_Button(250,688,90,25,"Refresh it");
	reapply->color(62);
	reapply->labelcolor(FL_BLACK);
	reapply->labelfont(3);	
	reapply->labelsize(15);
	reapply->callback(doRedraw);	
 
	/////////////////////////////////////////////////////////////////
		
	end_this = new Fl_Button(165,740, 80, 30, "Exit");
	end_this->color(201);		end_this->labelcolor(FL_BLACK);
	end_this->labelfont(3);		end_this->labelsize(18);	
	end_this->callback(exit_b);	

	opts_window->end(); 
	opts_window->show();
	/////////////////////////////////////////////////////////////////
	return(Fl::run());
   	return 0;
}
//////////////////////////////////////////////////////////////////////
/****    END_MAIN   END_MAIN   ENF_MAIN   END_MAIN   END_MAIN    ****/
//////////////////////////////////////////////////////////////////////
/****               Called Function Definition                   ****/
//////////////////////////////////////////////////////////////////////
/**
 The code follows the GRID line registration (default in GMT), 
 which means the nodes are centered on the grid line intersections.
 
 o------o------o------o
 |      |      |      |		Thus, the number of nodes are related
 |      |      |      |		to region and grid spacing by:
 o------o------o------o
 |      |      |      |		nx = (x_max-x_min)/x_inc + 1 = 4
 |      |      |      |		ny = (y_max-y_min)/y_inc + 1 = 3
 o------o------o------o 
 
 x_max, x_min, x_inc, and y_max, y_min, y_inc
 should be defined from data analysis.
 
 double x_min				Minimum x-value of region
 double x_max				Maximum x-value of region
 double x_inc				Node spacing in x-dimension
 
 double y_min				Minimum y-value of region
 double y_max				Maximum y-value of region
 double y_inc				Node spacing in y-dimension
 
 int nx						Number of nodes in the x-dimension
 int ny						Number of nodes in the y-dimension 
 
 double z_min				Minimum z-value in data set
 double z_max				Maximum z-value in data set
 
 float zdat[nx*ny]		z-values stored in scanline format
 
 Scanline format means that the data are stored in rows (y=const), 
 going from the “top” (y=y_max) to the “bottom” (y=y_min). 
 Data within each row are ordered from “left” (x=x_min)
 to “right” (x=x_max). 
 **/
void read_file_info(const char *name) // Reads ALL data into memory
{
	int i,j;
	char *ptr, buf[80], bigbuf[200], keys[]="1234567890-.";
	float xx,yy,zz,xold,yold,dx,dy;

	if ((in_fp = fopen(name, "r")) == NULL)
	{fl_message("ERROR opening input file!\nClose the window to exit");exit(1);}
	
	fseek(in_fp, 0L, SEEK_SET);	// go to beginning
	
	// Test if the first string contains non-digits	and count number of columns
	fgets(bigbuf,199,in_fp); 
	ptr = strtok (bigbuf," ");  j=0; 
	while (ptr != NULL)
	{
		sscanf(ptr,"%s",buf);
		if(strcspn(buf,keys) == 1)	//if(isdigit((int)buf[0]) == 0)
		{ fl_message("Input file contains non-decimal characters\n\
Close the window to exit."); fclose(in_fp); exit(1); }
		ptr = strtok (NULL, " ");	j++;	// j = column index
	}
	Ncur = j;
	if(Ncur != 3)
	{ fl_message("Input file should have three columns of data.\n\
Close the window to exit."); fclose(in_fp); exit(1); }
	
	///////////////////////////////////////////////////////////////
	//////////    Define the Range of input variables   ///////////
	///////////////////////////////////////////////////////////////
	
	fseek(in_fp, 0L, SEEK_SET);	// go to beginning and count number of lines
	
	x_min=y_min=z_min=FLT_MAX;  x_max=y_max=z_max=FLT_MIN;
	xold=yold=x_inc=y_inc=FLT_MAX/2.0;
	Ndat=0;
	while(fscanf(in_fp,"%f%f%f\n",&xx,&yy,&zz) != EOF)
	{
		if(xx>x_max) x_max=xx;  if(xx<x_min) x_min=xx;
		if(yy>y_max) y_max=yy;  if(yy<y_min) y_min=yy;
		if(zz>z_max) z_max=zz;  if(zz<z_min) z_min=zz;

		dx = fabsf(xx-xold);  dy = fabsf(yy-yold);
		if(dx<x_inc && dx>0.0) x_inc = dx;
		if(dy<y_inc && dy>0.0) y_inc = dy;	
		xold=xx; yold=yy;		
		Ndat++;
	}	
	nx = (int)((x_max-x_min)/x_inc + 1.000001);
	ny = (int)((y_max-y_min)/y_inc + 1.000001);
	nempty = nx*ny - Ndat;  // Number of empty nodes
	
	if(nempty<0)
	{ fl_message("ERROR in determining the range of input parameters.\n\
Close the window to exit."); fclose(in_fp); exit(1); }

	Zlow = z_min;	Zhigh = z_max;		// Default Z-values for mapping to colors
	pal = RGB_pal;						// Default color palette
	intercol=1;

	zdat = new (nothrow) float[nx * ny];
	if(zdat == 0)
	{ 
		fl_message("ERROR: NO memory to accomodate data!\n\
Close the window to exit...");	
		fclose(in_fp);  exit(1);
	}
	
	for(j=0;j<ny;j++)		// Empty nodes will be filled with NULLs
		for(i=0;i<nx;i++) zdat[nx*j+i] = 0.0; // Initialization with ZEROes

	///////////////////////////////////////////////////////////////
	////////////////    READ  DATA  into  MEMORY   ////////////////
	///////////////////////////////////////////////////////////////
	
	fseek(in_fp, 0L, SEEK_SET);		// Go to beginning of the file
									// and READ DATA into zdat[..]
	while(fscanf(in_fp,"%f%f%f\n",&xx,&yy,&zz) != EOF)
	{		
		i = (int)((xx-x_min+0.000001)/x_inc);
		j = (int)((yy-y_min+0.000001)/y_inc);
		if(i>=nx) 
		{ fl_message("ERROR reading X-data into memory.\n\
Close the window to exit"); fclose(in_fp); exit(1); }			
		if(j>=ny) 
		{ fl_message("ERROR reading Y-data into memory.\n\
Close the window to exit"); fclose(in_fp); exit(1); }
						 
		zdat[nx*j+i] = zz;
	}

	fclose(in_fp);

/**				Use this to Determne Screen Work Area

	fl_message("Screen Work Area in pixels:\n\
The height = %d, the width = %d\n\n\
Close the window to exit...",Fl::h(),Fl::w());			**/

}
////////////////////////////////////////////////////////////////
//		Below we assume that all the color palettes have the same
//		range of values from 0.0 to 1.0 with equal PALINC (==0.1)
//
void pal_interp(float palette[], float x, short colors[])
{					//			0.0 <= x <= 1.0 
	int i;
	short r0,g0,b0,dr,dg,db;
	float dx;
	
	r0=g0=b0=dr=dg=db=0;	
	i = (int)(x/PALINC + 0.000001);
	
	if(i<PALSZ && i>=0)		// should be always the case
	{
		dx = (x-palette[i])/PALINC;	
		r0 = palette[PALSZ+i];
		g0 = palette[PALSZ*2+i];
		b0 = palette[PALSZ*3+i];

		if(i<(PALSZ-1))
		{
			dr = (short)(dx*(palette[PALSZ+i+1] - r0));
			dg = (short)(dx*(palette[PALSZ*2+i+1] - g0));
			db = (short)(dx*(palette[PALSZ*3+i+1] - b0));		
		}	
	}
	
	colors[0] = r0+dr;
	colors[1] = g0+dg;
	colors[2] = b0+db;
}
//
void pal_discr(float palette[], float x, short colors[])
{							//	0.0 <= x <= 1.0 
	int i;

	i = (int)(x/PALINC + 0.000001);		//	0	<= i < palsz
	
	colors[0] = (short)palette[PALSZ+i];
	colors[1] = (short)palette[PALSZ*2+i];	
	colors[2] = (short)palette[PALSZ*3+i];	
}
